###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Initialize the server based on environment variables."""
import json
import logging
from importlib import resources
from importlib.metadata import entry_points

from django.db.models import Count

from .. import models as m
from ..plugin_manager import PluginManager
from .box import load_box_model
from .lang import load_lang_dst, load_lang_src
from .ocr import load_ocr_model
from .tsl import load_tsl_model

logger = logging.getLogger('ocr.general')

def init_most_used():
    """Initialize the server with the most used languages and models."""
    logger.info('Initializing server with most used languages and models')
    src = m.Language.objects.annotate(count=Count('trans_src')).order_by('-count').first()
    dst = m.Language.objects.annotate(count=Count('trans_dst')).order_by('-count').first()

    if src and src.count > 0:
        load_lang_src(src.iso1)
    if dst and dst.count > 0:
        load_lang_dst(dst.iso1)

    box = m.OCRBoxModel.objects.annotate(count=Count('box_runs')).order_by('-count').first()
    ocr = m.OCRModel.objects.annotate(count=Count('ocr_runs')).order_by('-count').first()
    tsl = m.TSLModel.objects.annotate(count=Count('tsl_runs')).order_by('-count').first()

    if box and box.count > 0:
        load_box_model(box.name)
    if ocr and ocr.count > 0:
        load_ocr_model(ocr.name)
    if tsl and tsl.count > 0:
        load_tsl_model(tsl.name)

def init_last_used():
    """Initialize the server with the most used languages and models."""
    logger.info('Initializing server with the last used languages and models')
    src = m.Language.get_last_loaded_src()
    dst = m.Language.get_last_loaded_dst()

    box = m.OCRBoxModel.get_last_loaded()
    ocr = m.OCRModel.get_last_loaded()
    tsl = m.TSLModel.get_last_loaded()

    if src:
        load_lang_src(src.iso1)
    else:
        logger.warning('No last source language found')
    if dst:
        load_lang_dst(dst.iso1)
    else:
        logger.warning('No last destination language found')
    if box:
        load_box_model(box.name)
    else:
        logger.warning('No last box model found')
    if ocr:
        load_ocr_model(ocr.name)
    else:
        logger.warning('No last OCR model found')
    if tsl:
        load_tsl_model(tsl.name)
    else:
        logger.warning('No last TSL model found')

def auto_create_languages():
    """Create Language objects from json file."""
    lang_file = resources.files('ocr_translate.ocr_tsl').joinpath('languages.json')
    with lang_file.open(encoding='utf-8') as f:
        langs = json.load(f)

    for lang in langs:
        logger.debug(f'Creating language: {lang}')
        name = lang.pop('name')
        iso1 = lang.pop('iso1')
        iso2t = lang.pop('iso2t')
        iso2b = lang.pop('iso2b')
        iso3 = lang.pop('iso3')
        def_opt = lang.pop('default_options', {})
        opt_obj, _ = m.OptionDict.objects.get_or_create(options=def_opt)
        l, _ = m.Language.objects.get_or_create(name=name, iso1=iso1, iso2t=iso2t, iso2b=iso2b, iso3=iso3)
        l.default_options = opt_obj
        # for k,v in lang.items():
        #     setattr(l, k, v)
        l.save()

    m.OptionDict.objects.get_or_create(options={})

def load_ept_data(namespace):
    """Load all entrypoints from a namespace into a list"""
    return[_.load() for _ in entry_points(group=namespace)]

# Pop and set after so that running this after a migration should modify the existing model (with less
# attributes) instead of creating a new one
def add_box_model(ep_dict: dict) -> m.OCRBoxModel:
    """Create OCRBoxModel object from dict."""
    ep_dict = ep_dict.copy()
    logger.debug(f'Creating box model: {ep_dict}')
    lang = ep_dict.pop('lang')
    lcode = ep_dict.pop('lang_code')
    entrypoint = ep_dict.pop('entrypoint')
    iso1_map = ep_dict.pop('iso1_map', {})
    def_opt = ep_dict.pop('default_options', {})
    opt_obj, _ = m.OptionDict.objects.get_or_create(options=def_opt)
    model, _ = m.OCRBoxModel.objects.get_or_create(**ep_dict)
    model.default_options = opt_obj
    model.entrypoint = entrypoint
    model.language_format = lcode
    model.iso1_map = iso1_map
    model.languages.clear()
    for l in lang:
        model.languages.add(m.Language.objects.get(iso1=l))
    model.save()
    return model

def add_ocr_model(ep_dict: dict) -> m.OCRModel:
    """Create OCRModel object from dict."""
    ep_dict = ep_dict.copy()
    logger.debug(f'Creating ocr model: {ep_dict}')
    lang = ep_dict.pop('lang')
    lcode = ep_dict.pop('lang_code')
    ocr_mode = ep_dict.pop('ocr_mode', m.OCRModel.MERGED)
    entrypoint = ep_dict.pop('entrypoint')
    iso1_map = ep_dict.pop('iso1_map', {})
    def_opt = ep_dict.pop('default_options', {})
    opt_obj, _ = m.OptionDict.objects.get_or_create(options=def_opt)
    model, _ = m.OCRModel.objects.get_or_create(**ep_dict)
    model.default_options = opt_obj
    model.language_format = lcode
    model.ocr_mode = ocr_mode
    model.iso1_map = iso1_map
    model.entrypoint = entrypoint
    model.languages.clear()
    for l in lang:
        model.languages.add(m.Language.objects.get(iso1=l))
    model.save()
    return model

def add_tsl_model(ep_dict: dict) -> m.TSLModel:
    """Create TSLModel object from dict."""
    ep_dict = ep_dict.copy()
    logger.debug(f'Creating tsl model: {ep_dict}')
    src = ep_dict.pop('lang_src', [])
    dst = ep_dict.pop('lang_dst', [])
    lcode = ep_dict.pop('lang_code', None)
    entrypoint = ep_dict.pop('entrypoint', None)
    iso1_map = ep_dict.pop('iso1_map', {})
    def_opt = ep_dict.pop('default_options', {})
    opt_obj, _ = m.OptionDict.objects.get_or_create(options=def_opt)
    model, _ = m.TSLModel.objects.get_or_create(**ep_dict)
    model.default_options = opt_obj
    model.language_format = lcode
    model.iso1_map = iso1_map
    model.entrypoint = entrypoint
    model.src_languages.clear()
    for l in src:
        logger.debug(f'Adding src language: {l}')
        kwargs = {lcode: l}
        model.src_languages.add(*m.Language.objects.filter(**kwargs))

    model.dst_languages.clear()
    for l in dst:
        logger.debug(f'Adding dst language: {l}')
        kwargs = {lcode: l}
        model.dst_languages.add(*m.Language.objects.filter(**kwargs))
    model.save()
    return model

def ensure_plugins():
    """Ensure that all plugins are installed on initialization.
    This is used to make sure that running the server with a new DEVICE will have the correct dependencies installed."""
    logger.info('Ensuring that all plugins are loaded')
    pmng = PluginManager()
    known = set(_['name'] for _ in pmng.plugins_data)
    installed = set(pmng.plugins)
    for plugin in known:
        if plugin in installed:
            pmng.install_plugin(plugin)



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Caches for list of models to be sent to the frontend."""
from django.db.models import Count
from django.db.models.signals import post_save
from django.dispatch import receiver

from .. import models as m
from .lang import get_lang_dst, get_lang_src
from .signals import refresh_model_cache_signal

# Caches for the languages ordered by number of translations
ALL_LANG_SRC: list[m.Language] = None
ALL_LANG_DST: list[m.Language] = None

ALLOWED_BOX_MODELS: list[m.OCRBoxModel] = None
ALLOWED_OCR_MODELS: list[m.OCRModel] = None
ALLOWED_TSL_MODELS: list[m.TSLModel] = None

def refresh_model_cache():
    """Refresh the models cached entry."""
    global ALLOWED_BOX_MODELS
    global ALLOWED_OCR_MODELS
    global ALLOWED_TSL_MODELS

    lang_src = get_lang_src()
    lang_dst = get_lang_dst()

    ALLOWED_BOX_MODELS = []
    ALLOWED_OCR_MODELS = []
    ALLOWED_TSL_MODELS = []
    if not lang_src is None:
        ALLOWED_BOX_MODELS = m.OCRBoxModel.objects.annotate(
            count=Count('box_runs')
            ).filter(languages=lang_src, active=True).order_by('-count').all()
        ALLOWED_OCR_MODELS = m.OCRModel.objects.annotate(
            count=Count('ocr_runs')
            ).filter(languages=lang_src, active=True).order_by('-count').all()
        if not lang_dst is None:
            ALLOWED_TSL_MODELS = m.TSLModel.objects.annotate(
                count=Count('tsl_runs')
                ).filter(src_languages=lang_src, dst_languages=lang_dst, active=True).order_by('-count').all()

def refresh_lang_cache():
    """Refresh the languages cached entry."""
    global ALL_LANG_SRC
    global ALL_LANG_DST
    ALL_LANG_SRC = m.Language.objects.annotate(count=Count('trans_src')).order_by('-count').all()
    ALL_LANG_DST = m.Language.objects.annotate(count=Count('trans_dst')).order_by('-count').all()

def get_all_lang_src():
    """Return all the source languages."""
    if ALL_LANG_SRC is None:
        refresh_lang_cache()
    return ALL_LANG_SRC

def get_all_lang_dst():
    """Return all the destination languages."""
    if ALL_LANG_DST is None:
        refresh_lang_cache()
    return ALL_LANG_DST

def get_allowed_box_models():
    """Return the allowed box models."""
    if ALLOWED_BOX_MODELS is None:
        refresh_model_cache()
    return ALLOWED_BOX_MODELS

def get_allowed_ocr_models():
    """Return the allowed OCR models."""
    if ALLOWED_OCR_MODELS is None:
        refresh_model_cache()
    return ALLOWED_OCR_MODELS

def get_allowed_tsl_models():
    """Return the allowed TSL models."""
    if ALLOWED_TSL_MODELS is None:
        refresh_model_cache()
    return ALLOWED_TSL_MODELS

@receiver(post_save, sender=m.Language)
def refres_lang_callback(sender, instance, **kwargs): # pylint: disable=unused-argument
    """Callback to refresh the cached language list when a language is added/modified."""
    refresh_lang_cache()

@receiver(refresh_model_cache_signal)
@receiver(post_save, sender=m.OCRBoxModel)
@receiver(post_save, sender=m.OCRModel)
@receiver(post_save, sender=m.TSLModel)
def refresh_models_callback(sender, **kwargs): # pylint: disable=unused-argument
    """Callback to refresh the cached model list when a language is added/modified."""
    refresh_model_cache()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Manages the globaly selected source and destination language."""
from .. import models as m
from .signals import refresh_model_cache_signal

LANG_SRC: m.Language = None
LANG_DST: m.Language = None

def get_lang_src():
    """Return the source language."""
    return LANG_SRC

def get_lang_dst():
    """Return the destination language."""
    return LANG_DST

def load_lang_src(iso1):
    """Load the source language."""
    global LANG_SRC
    LANG_SRC = m.Language.objects.get(iso1=iso1)
    LANG_SRC.load_src()
    refresh_model_cache_signal.send(sender=None)

def load_lang_dst(iso1):
    """Load the destination language."""
    global LANG_DST
    LANG_DST = m.Language.objects.get(iso1=iso1)
    LANG_DST.load_dst()
    refresh_model_cache_signal.send(sender=None)



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django views for the ocr_translate app."""
# pylint: disable=too-many-return-statements
import base64
import hashlib
import io
import logging
import traceback
from typing import Union

import numpy as np
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.middleware import csrf
from django.views.decorators.csrf import csrf_exempt
from PIL import Image

from . import __version__array__
from . import models as m
from .entrypoint_manager import ep_manager
from .ocr_tsl.box import get_box_model, load_box_model, unload_box_model
from .ocr_tsl.cached_lists import (get_all_lang_dst, get_all_lang_src,
                                   get_allowed_box_models,
                                   get_allowed_ocr_models,
                                   get_allowed_tsl_models)
from .ocr_tsl.full import ocr_tsl_pipeline_lazy, ocr_tsl_pipeline_work
from .ocr_tsl.lang import (get_lang_dst, get_lang_src, load_lang_dst,
                           load_lang_src)
from .ocr_tsl.ocr import get_ocr_model, load_ocr_model, unload_ocr_model
from .ocr_tsl.tsl import get_tsl_model, load_tsl_model, unload_tsl_model
from .plugin_manager import PluginManager
from .queues import main_queue as q
from .request_decorators import (get_backend_langs, get_backend_models,
                                 get_data_deserializer, method_or_405,
                                 post_data_deserializer, use_lock,
                                 wait_for_lock)
from .tries import load_trie_src

logger = logging.getLogger('ocr.general')

PMNG = PluginManager()


@method_or_405(['GET'])
def handshake(request: HttpRequest) -> JsonResponse:
    """Handshake with the client."""
    csrf.get_token(request)

    logger.debug(f'Handshake: {str(get_ocr_model())}, {str(get_tsl_model())}')
    lang_src = get_lang_src()
    lang_dst = get_lang_dst()

    languages = get_all_lang_src()
    languages_src = get_all_lang_src()
    languages_dst = get_all_lang_dst()

    box_models = get_allowed_box_models()
    ocr_models = get_allowed_ocr_models()
    tsl_models = get_allowed_tsl_models()

    box_model = get_box_model() or ''
    ocr_model = get_ocr_model() or ''
    tsl_model = get_tsl_model() or ''

    lang_src = getattr(lang_src, 'iso1', None) or ''
    lang_dst = getattr(lang_dst, 'iso1', None) or ''

    res = JsonResponse({
        'version': __version__array__,
        'Languages': [_.iso1 for _ in languages_src],
        'Languages_src': [_.iso1 for _ in languages_src],
        'Languages_dst': [_.iso1 for _ in languages_dst],
        'Languages_hr': [_.name for _ in languages],
        'BOXModels': [str(_) for _ in box_models],
        'OCRModels': [str(_) for _ in ocr_models],
        'TSLModels': [str(_) for _ in tsl_models],

        'box_selected': str(box_model),
        'ocr_selected': str(ocr_model),
        'tsl_selected': str(tsl_model),
        'lang_src': lang_src,
        'lang_dst': lang_dst,
        })

    # res['Access-Control-Allow-Origin'] = request.headers.get('Origin', '*')
    # res['Access-Control-Allow-Credentials'] = 'true'

    return res

@csrf_exempt
@method_or_405(['POST'])
@post_data_deserializer(['box_model_id', 'ocr_model_id', 'tsl_model_id'], required=False)
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def set_models(request: HttpRequest, box_model_id, ocr_model_id, tsl_model_id) -> JsonResponse:
    """Handle a POST request to load models.
    Expected data:
    {
        'box_model_id': 'id',
        'ocr_model_id': 'id',
        'tsl_model_id': 'id',
    }
    """
    logger.info(f'LOAD MODELS: {box_model_id}, {ocr_model_id}, {tsl_model_id}')

    try:
        if not box_model_id is None and not box_model_id == '':
            load_box_model(box_model_id)
        if not ocr_model_id is None and not ocr_model_id == '':
            load_ocr_model(ocr_model_id)
        if not tsl_model_id is None and not tsl_model_id == '':
            load_tsl_model(tsl_model_id)
    except Exception as exc:
        logger.error(f'Failed to load models: {exc}')
        return JsonResponse({'error': str(exc)}, status=400)

    return JsonResponse({})

@csrf_exempt
@method_or_405(['POST'])
@post_data_deserializer(['lang_src', 'lang_dst'], required=True)
def set_lang(request: HttpRequest, lang_src, lang_dst) -> JsonResponse:
    """Handle a POST request to set languages.
    Expected data:
    {
        'lang_src': 'id',
        'lang_dst': 'id',
    }
    """
    logger.info(f'SET LANG: {lang_src}, {lang_dst}')

    old_src = get_lang_src()
    old_dst = get_lang_dst()
    try:
        load_lang_src(lang_src)
        load_lang_dst(lang_dst)
        load_trie_src(lang_src)
        # load_trie_dst(lang_dst)
    except Exception as exc:
        return JsonResponse({'error': str(exc)}, status=400)
    new_src = get_lang_src()
    new_dst = get_lang_dst()

    check1 = old_src != new_src
    check2 = old_dst != new_dst
    if check1 or check2:
        tsl_model = get_tsl_model()
        if (
            not tsl_model is None and
            (
                new_src not in tsl_model.src_languages.all() or
                new_dst not in tsl_model.dst_languages.all()
            )):
            unload_tsl_model()
    if check1:
        box_model = get_box_model()
        ocr_model = get_ocr_model()
        if not box_model is None and (new_src not in box_model.languages.all()):
            unload_box_model()
        if not ocr_model is None and (new_src not in ocr_model.languages.all()):
            unload_ocr_model()
    # if check2:
    #     pass

    return JsonResponse({})

@csrf_exempt
@method_or_405(['POST'])
@post_data_deserializer(['text'], required=True)
@get_backend_langs(strict=True)
@get_backend_models(strict=True)
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def run_tsl(request: HttpRequest, text, tsl_model: m.TSLModel, **kwargs) -> JsonResponse:
    """Handle a POST request to run translation.
    Expected data:
    {
        'text': 'text',
    }
    """
    src_obj, _ = m.Text.objects.get_or_create(text=text)
    dst_obj = tsl_model.translate(src_obj, get_lang_src(), get_lang_dst())
    dst_obj = next(dst_obj)

    return JsonResponse({
        'text': dst_obj.text,
        })

@method_or_405(['GET'])
@get_backend_langs(strict=True)
@get_backend_models(strict=True)
@get_data_deserializer(['text'], required=True)
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def run_tsl_get_xunityautotrans(
    request: HttpRequest, tsl_model: m.TSLModel, text: str,
    lang_src: m.Language, lang_dst: m.Language, **kwargs
    ) -> JsonResponse:
    """Handle a GET request to run translation.
    Expected parameters:
    {
        'text': 'text',
    }
    """
    src_obj, _ = m.Text.objects.get_or_create(text=text)
    dst_obj = tsl_model.translate(src_obj, lang_src, lang_dst)
    dst_obj = next(dst_obj)

    return HttpResponse(dst_obj.text)

@csrf_exempt
@method_or_405(['POST'])
@get_backend_langs(strict=True)
@get_backend_models(strict=True)
@post_data_deserializer(['contents', 'md5', 'force', 'options'], required=False)
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def run_ocrtsl(  # pylint: disable=too-many-locals
    request: HttpRequest,
    lang_src: m.Language, lang_dst: m.Language,
    box_model: m.OCRBoxModel, ocr_model: m.OCRModel, tsl_model: m.TSLModel,
    contents: str, md5: str, force: bool, options: dict,
    ) -> JsonResponse:
    """Handle a POST request to run OCR and translation.
    Expected data:
    {
        'contents': 'base64',
        'md5': 'md5',
        'force': 'bool',
        'options': 'dict',
    }
    """
    b64 = contents
    frc = force
    opt = options or {}

    opt_box = opt.get(box_model.name if box_model else None, {})
    opt_ocr = opt.get(ocr_model.name if ocr_model else None, {})
    opt_tsl = opt.get(tsl_model.name if tsl_model else None, {})

    options_box, _ = m.OptionDict.objects.get_or_create(options=opt_box)
    options_ocr, _ = m.OptionDict.objects.get_or_create(options=opt_ocr)
    options_tsl, _ = m.OptionDict.objects.get_or_create(options=opt_tsl)

    if b64 is None:
        logger.info('No contents, trying to lazyload')
        if frc:
            return JsonResponse({'error': 'Cannot force ocr without contents'}, status=400)
        try:
            res = ocr_tsl_pipeline_lazy(
                md5,
                options_box=options_box,
                options_ocr=options_ocr,
                options_tsl=options_tsl,
                )
        except ValueError:
            logger.info('Failed to lazyload ocr')
            return JsonResponse({'error': 'Failed to lazyload ocr'}, status=406)
    else:
        binary = base64.b64decode(b64)
        # Doing md5 on the base64 to have consistency with the JS generate one
        # Can't find a way to run md5 on the binary in JS (the blob does not work)
        if md5 != hashlib.md5(b64.encode('utf-8')).hexdigest():
            return JsonResponse({'error': 'md5 mismatch'}, status=400)
        logger.debug(f'md5 {md5} <- {len(binary)} bytes')

        img = Image.open(io.BytesIO(binary))
        # Needed to make sure the image is loaded synchronously before going forward
        # Enforce thread safety. Maybe there is a way to do it without numpy?
        np.array(img)

        # Check if same request is already in queue. If yes attach listener to it
        id_ = (
            md5,
            lang_src.id, lang_dst.id,
            box_model.id, ocr_model.id, tsl_model.id,
            options_box.id, options_ocr.id, options_tsl.id,
            )

        msg = q.put(
            id_ = id_,
            msg = {
                'args': (img, md5),
                'kwargs': {
                    'force': frc,
                    # 'options': opt,
                    'options_box': options_box,
                    'options_ocr': options_ocr,
                    'options_tsl': options_tsl,
                    },
            },
            handler = ocr_tsl_pipeline_work,
        )

        res = msg.response()

        if isinstance(res, Exception):
            logger.error(f'Failed to run ocr: {res}')
            logger.debug(traceback.print_exception(type(res), res, res.__traceback__))
            return JsonResponse({'error': str(res)}, status=500)


    return JsonResponse({
        'result': res,
        })


@csrf_exempt
@method_or_405(['GET'])
@get_backend_langs(strict=True)
@get_data_deserializer(['text'], required=True)
def get_translations(
    request: HttpRequest,
    lang_src: m.Language, lang_dst: m.Language,
    text: str,
    ) -> JsonResponse:
    """Handle a GET request to get translations.
    Expected parameters:
    {
        'text': 'text',
    }
    """
    text_obj = m.Text.objects.filter(text=text).first()
    if text_obj is None:
        return JsonResponse({'error': 'text not found'}, status=404)

    translations = text_obj.to_trans.filter(
        lang_src=lang_src,
        lang_dst=lang_dst,
        )
    return JsonResponse({
        'translations': [{
            'model': str(_.model),
            'text': _.result.text,
            } for _ in translations],
        })

@csrf_exempt
@method_or_405(['POST'])
@get_backend_langs(strict=True)
@post_data_deserializer(['text', 'translation'], required=True)
def set_manual_translation(
    request: HttpRequest,
    lang_src: m.Language, lang_dst: m.Language,
    text: str, translation: str,
    ) -> JsonResponse:
    """Handle a POST request to apply a manual translation.
    Expected data:
    {
        'text': 'text',
        'translation': 'translation',
    }
    """
    text_obj = m.Text.objects.filter(text=text).first()
    if text_obj is None:
        return JsonResponse({'error': 'text not found'}, status=404)

    manual_model = m.TSLModel.objects.filter(name='manual').first()
    params = {
        'model': manual_model,
        'text': text_obj,
        'lang_src': lang_src,
        'lang_dst': lang_dst,
        'options': m.OptionDict.objects.get(options={}),
    }

    tsl_run_obj = m.TranslationRun.objects.filter(**params).first()
    if tsl_run_obj is None:
        res_obj, _ = m.Text.objects.get_or_create(text=translation)
        params['result'] = res_obj
        tsl_run_obj = m.TranslationRun.objects.create(**params)
    else:
        tsl_run_obj.result.text = translation
        tsl_run_obj.result.save()

    return JsonResponse({})

def get_default_options_from_cascade(
        objects: list[Union[str, 'm.OptionDict']], option: str, default: Union[int, float, str, bool] = None
        ) -> Union[int, float, str, bool]:
    """Get the default value of an option from a cascade of objects.

    Args:
        objects (list[Union[str, OptionDict]]): List of option objects or string identifying a model type.
        option (str): Name of the option to get.
        default (Union[int, float, str, bool], optional): Default value to return for the option. Defaults to None.

    Raises:
        ValueError: Invalid string in objects list.
        TypeError: Invalid type in objects list.

    Returns:
        Union[int, float, str, bool]: The default value of the option.
    """
    res = default
    for obj in objects:
        if obj is None:
            continue
        if isinstance(obj, m.OptionDict):
            res = obj.options.get(option, res)
        elif isinstance(obj, str):
            if obj == 'lang_src':
                model = get_lang_src()
            elif obj == 'lang_dst':
                model = get_lang_dst()
            elif obj == 'box_model':
                model = get_box_model()
            elif obj == 'ocr_model':
                model = get_ocr_model()
            elif obj == 'tsl_model':
                model = get_tsl_model()
            else:
                raise ValueError(f'Unknown option cascade object: {obj}')
            res = model.default_options.options.get(option, res)
        else:
            raise TypeError(f'Cannot get default options from {type(obj)}')
    return res

@method_or_405(['GET'])
@get_backend_models(strict=False)
@get_data_deserializer([], required=False)
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def get_active_options(
    request: HttpRequest,
    box_model: m.OCRBoxModel, ocr_model: m.OCRModel, tsl_model: m.TSLModel,
    ) -> JsonResponse:
    """Handle a GET request to get active options."""
    res = {}
    for typ,model in [
        ('box_model', box_model),
        ('ocr_model', ocr_model),
        ('tsl_model', tsl_model),
        ]:
        ptr = res.setdefault(typ, {})
        if model == '':
            continue

        for opt,val in model.ALLOWED_OPTIONS.items():
            val = val.copy()
            ptr[opt] = val
            default = val['default']
            if isinstance(default, tuple):
                if default[0] == 'cascade':
                    default = get_default_options_from_cascade(default[1], opt, default[2])
                    default = val['type'](default)
                    val['default'] = default
                else:
                    raise ValueError(f'Unknown default action type: {default[0]}')
            val['type'] = val['type'].__name__

    return JsonResponse({'options': res})

@method_or_405(['GET'])
@wait_for_lock('plugin')
@use_lock('block_plugin_changes', blocking=False)
def get_plugin_data(request: HttpRequest) -> JsonResponse:
    """Handle a GET request to get plugins."""
    resp = {}
    tpl = {
        'homepage': None,
        'warning': None,
        'description': None,
    }
    for plugin in PMNG.plugins_data:
        ptr = {**tpl, **plugin}
        name = ptr.pop('name')
        ptr.pop('dependencies', None)
        ptr['installed'] = name in PMNG.plugins
        resp[name] = ptr
    return JsonResponse(resp)

@csrf_exempt
@method_or_405(['POST'])
@post_data_deserializer(['plugins'], required=True)
@use_lock('plugin', blocking=True)
@use_lock('block_plugin_changes', blocking=True)
def manage_plugins(request: HttpRequest, plugins: dict[str, bool]) -> JsonResponse:
    """Handle a POST request to install a plugin."""
    logger.debug(f'Manage plugins: {plugins}')
    with ep_manager():
        for plugin, present in plugins.items():
            if present:
                PMNG.install_plugin(plugin)
            else:
                PMNG.uninstall_plugin(plugin)
    return JsonResponse({})



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Full OCR + translation pipelines."""
import logging

from PIL import Image

from .. import models as m
from .box import get_box_model
from .lang import get_lang_dst, get_lang_src
from .ocr import get_ocr_model
from .tsl import get_tsl_model

logger = logging.getLogger('ocr.general')

def ocr_tsl_pipeline_lazy(
        md5: str,
        options_box: m.OptionDict,
        options_ocr: m.OptionDict,
        options_tsl: m.OptionDict,
        ) -> list[dict]:
    """
    Try to lazily generate reponse from md5.
    Should raise a ValueError if the operation is not possible (fails at any step).
    """
    box_model = get_box_model()
    ocr_model = get_ocr_model()
    tsl_model = get_tsl_model()

    favor_manual = options_tsl.options.get('favor_manual', True)
    logger.debug(f'LAZY: START {md5}')
    res = []
    try:
        img_obj= m.Image.objects.get(md5=md5)
    except m.Image.DoesNotExist as exc:
        raise ValueError(f'Image with md5 {md5} does not exist') from exc
    _, bbox_obj_list = box_model.box_detection(img_obj, get_lang_src(), options=options_box)

    for bbox_obj in bbox_obj_list:
        text_obj = ocr_model.ocr(bbox_obj, get_lang_src(), options=options_ocr)
        text_obj = next(text_obj)

        tsl_obj = tsl_model.translate(
            text_obj, get_lang_src(), get_lang_dst(),
            options=options_tsl, favor_manual=favor_manual,
            lazy=True
            )
        tsl_obj = next(tsl_obj)

        text = text_obj.text
        new = tsl_obj.text

        res.append({
            'ocr': text,
            'tsl': new,
            'box': bbox_obj.lbrt,
            })

    logger.debug('LAZY: DONE')
    return res

# This is already kinda lazy, but the idea for the lazy version is to
# check if all results are available just with the md5, and if not,
# ask the extension to send the binary to minimize traffic
def ocr_tsl_pipeline_work(  # pylint: disable=too-many-locals
        img: Image.Image, md5: str,
        options_box: m.OptionDict,
        options_ocr: m.OptionDict,
        options_tsl: m.OptionDict,
        force: bool = False,
        ) -> list[dict]:
    """
    Generate response from md5 and binary.
    Will attempt to behave lazily at every step unless force is True.
    """
    box_model = get_box_model()
    ocr_model = get_ocr_model()
    tsl_model = get_tsl_model()
    lang_src = get_lang_src()
    lang_dst = get_lang_dst()

    favor_manual = options_tsl.options.get('favor_manual', True)
    logger.debug(f'WORK: START {md5}')

    img_obj, _ = m.Image.objects.get_or_create(md5=md5)
    bbox_obj_list_single, bbox_obj_list_merged = box_model.box_detection(
        img_obj, lang_src ,image=img, options=options_box
        )

    bbox_obj_list = bbox_obj_list_merged
    if ocr_model.ocr_mode == ocr_model.SINGLE:
        bbox_obj_list = bbox_obj_list_single

    texts = []
    for bbox_obj in bbox_obj_list:
        logger.debug(str(bbox_obj))

        text_obj = ocr_model.ocr(bbox_obj, lang_src, image=img, force=force, block=False, options=options_ocr)
        next(text_obj)
        texts.append(text_obj)

    texts = [next(_) for _ in texts]
    # If OCR is running in single element mode, must merge the results and pretend runs on the merged
    # bounding boxes were done with the merged text as outputs
    if ocr_model.ocr_mode == ocr_model.SINGLE:
        logger.debug(f'OCR DONE (single): {texts}')
        str_list = [_.text for _ in texts]

        merged_text = ocr_model.merge_single_result(
            lang_src.iso1,
            str_list,
            bbox_obj_list_single,
            bbox_obj_list_merged
            )
        texts = []
        for text, bbox_obj in zip(merged_text, bbox_obj_list_merged):
            text_obj, _ = m.Text.objects.get_or_create(text=text)
            params = {
                'bbox': bbox_obj,
                'model': ocr_model,
                'lang_src': lang_src,
                'options': options_ocr,
                'result_merged': text_obj
            }
            merged_run = m.OCRRun.objects.create(**params)

            # Pretend the current texts are the merged ones
            texts.append(merged_run.result_merged)

            logger.debug(f'OCR DONE (mock_merged): {text}')


    logger.debug(f'OCR DONE: {texts}')

    trans = []
    for text_obj in texts:
        tsl_obj = tsl_model.translate(
            text_obj, lang_src, lang_dst, options=options_tsl,
            force=force, block=False, favor_manual=favor_manual
            )
        next(tsl_obj)
        trans.append(tsl_obj)

    trans = [next(_) for _ in trans]
    logger.debug(f'TRANSLATION DONE: {trans}')

    res = []
    for bbox_obj, text_obj, tsl_obj in zip(bbox_obj_list_merged, texts, trans):
        text = text_obj.text
        new = tsl_obj.text

        res.append({
            'ocr': text,
            'tsl': new,
            'box': bbox_obj.lbrt,
            })

    logger.debug('WORK: DONE')
    return res



# Generated by Django 4.2.3 on 2023-12-17 20:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0010_rename_from_ocr_bbox_from_ocr_merged_squashed_0016_alter_ocrrun_result_merged'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ocrboxmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ocrmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='tslmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
    ]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Manages changes in entrypoints after installation/removal of plugins."""

import contextlib
import logging
from importlib.metadata import EntryPoint, entry_points

from . import models as m
from .ocr_tsl import initializers as ini
from .ocr_tsl.box import get_box_model, unload_box_model
from .ocr_tsl.ocr import get_ocr_model, unload_ocr_model
from .ocr_tsl.signals import refresh_model_cache_signal
from .ocr_tsl.tsl import get_tsl_model, unload_tsl_model

logger = logging.getLogger('ocr.general')

def get_group_entrypoints(group: str) -> set[EntryPoint]:
    """Get all entrypoints for a group"""
    return set(ep for ep in entry_points(group=group))

GROUPS = {
    'ocr_translate.box_data': (m.OCRBoxModel, ini.add_box_model, get_box_model, unload_box_model),
    'ocr_translate.ocr_data': (m.OCRModel, ini.add_ocr_model, get_ocr_model, unload_ocr_model),
    'ocr_translate.tsl_data': (m.TSLModel, ini.add_tsl_model, get_tsl_model, unload_tsl_model),
}

@contextlib.contextmanager
def ep_manager():
    """Context manager for entrypoint management.
    Provides a context that monitors changes to server server specific entrypoints.
    Make sures models from added entrypoints are created/activated and models from removed
    entrypoints are deactivated.
    """
    before: dict[str, set] = {}
    for grp in GROUPS:
        before[grp] = get_group_entrypoints(grp)

    yield

    after: dict[str, set] = {}
    for grp in GROUPS:
        after[grp] = get_group_entrypoints(grp)

    flag = False
    for grp, data in GROUPS.items():
        cls, create_func, get_func, unload_func = data
        loaded_model = get_func()
        added = after[grp] - before[grp]
        # Using after to ensure that when using same plugin folder with new db, the models are created
        for ept in after[grp]:
            if ept in added:
                logger.info(f'New entrypoint {ept.name} found')
            data = ept.load()
            model_id = data['name']
            try:
                model = cls.objects.get(name=model_id)
            except cls.DoesNotExist:
                model = create_func(data.copy())
            model.active = True
            model.save()

        removed = before[grp] - after[grp]
        for ept in removed:
            data = ept.load()
            model_id = data['name']
            try:
                model = cls.objects.get(name=model_id)
                if loaded_model is not None:
                    if loaded_model.name == model.name:
                        unload_func()
                model.active = False
                model.save()
                logger.info(f'Entrypoint {ept.name} removed')
            except cls.DoesNotExist:
                logger.warning(f'Could not find model {model_id} to deactivate')
                continue

        flag = flag or bool(added or removed)

    if flag:
        refresh_model_cache_signal.send(sender=None)



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Utils to manage ocr_translate plugins."""

# pylint: disable=redefined-outer-name,unspecified-encoding,too-many-branches

import functools
import json
import logging
import os
import platform
import shutil
import site
import subprocess
import sys
import time
from importlib import resources
from pathlib import Path
from threading import Lock

from django.apps import apps
from django.conf import settings

GENERIC_SCOPE = 'generic'
DEFAULT_OCT_BASE_DIR: Path = Path.home() / '.ocr_translate'

logger = logging.getLogger('ocr.general')

def find_site_packages(start_dir: Path) -> Path:
    """Find the site-packages directory."""
    for pth in start_dir.rglob('site-packages'):
        return pth
    return None

def reload_django_apps():
    """Reload the plugins."""
    apps.app_configs = {}
    apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
    apps.clear_cache()
    apps.populate(settings.INSTALLED_APPS)

def get_safe_name(name: str) -> str:
    """Get a safe name for the plugin."""
    return name.replace('-', '_min_')

def install_overrides_decorator(func):
    """Decorator to install a package."""
    arg_names = ['version', 'extras', 'scope', 'system']

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = list(args)
        cls = args.pop(0)
        if args:
            if 'name' in kwargs:
                raise TypeError('Got multiple values for argument name')
            name = args.pop(0)
        elif 'name' in kwargs:
            name = kwargs.pop('name')
        else:
            raise ValueError('Name required')

        new_args = {'name': name}
        for val,key in zip(args, arg_names):
            new_args[key] = val
        for key, val in kwargs.items():
            if key in new_args:
                raise  TypeError(f'Got multiple values for argument {key}')
            new_args[key] = val

        safe_name = get_safe_name(name).upper()
        for key in arg_names:
            over_name = f'OCT_PKG_{safe_name}_{key.upper()}'
            if over_name in os.environ:
                logger.debug(f'Overriding {key} for {name} with {os.environ[over_name]}')
                new_args[key] = os.environ[over_name]

        return func(cls, **new_args)

    return wrapped

def _pip_install(package: str, prefix: str, extras: list[str] = None, retries: int = 3):
    """Install a package with pip."""
    if extras is None:
        extras = []
    logger.info(f'Installing {package}')
    cmd = [
        'pip', 'install', package,
        '--ignore-installed',
        '--no-deps',
        '--no-build-isolation',
        f'--prefix={prefix}',
        ] + extras
    attempts = 0
    while True:
        attempts += 1
        try:
            res = subprocess.run(cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as exc:
            if attempts <= retries:
                logger.error(
                    f'Error installing {package}, retrying {attempts}/{retries}. '
                    'Waiting 5s before retrying...'
                    )
                time.sleep(5)
                continue
            logger.error(f'Error installing {package}')
            logger.error(exc.stdout.decode())
            logger.error(exc.stderr.decode())
            raise exc
        break
    logger.debug(res.stdout.decode())

class PluginManager:  # pylint: disable=too-many-instance-attributes
    """Manage the plugins."""
    _SINGLETON = None

    device: str = None
    scopes: list[str] = None

    base_dir: Path = None
    plugin_dir: Path = None
    PLUGIN_SP: dict[str, Path] = {}

    plugin_list_file: Path = None
    installed_file: Path = None

    def __new__(cls, *args, **kwargs):
        if not cls._SINGLETON:
            cls._SINGLETON = super().__new__(cls, *args, **kwargs)
        return cls._SINGLETON

    def __init__(self):
        """Initialize the plugin manager."""
        self.device = os.environ.get('DEVICE', 'cpu')
        self.scopes = [GENERIC_SCOPE, self.device]
        self.base_dir = Path(os.environ.get('OCT_BASE_DIR', DEFAULT_OCT_BASE_DIR))
        self.plugin_dir = self.base_dir / 'plugins'

        self.plugin_list_file = self.base_dir / 'plugins.json'
        self.installed_file = self.plugin_dir / 'installed.json'

        self.disabled = os.environ.get('OCT_DISABLE_PLUGINS', False)

        self.lock_plugin = Lock()
        self.lock_pkg = Lock()

        self._plugins = None
        self._plugin_data = None
        self._installed = None
        self.system = platform.system().lower()

        self.ensure_pip()
        self.initialize_scopes()

        # Ensure no previous empty or wrong site-packages dirs are present
        while (tmp_dir := find_site_packages(self.plugin_dir)) is not None:
            shutil.rmtree(tmp_dir)

    def ensure_pip(self):
        """Ensure pip is installed."""
        if self.disabled:
            return
        try:
            subprocess.run(['pip', '-V'], capture_output=True, check=True)
        except FileNotFoundError as exc:
            msg = ' ---- '.join([
                '!!!!!'
                'pip not found, please install a version of python https://www.python.org/downloads/'
                'Make sure to check the `Add python.exe to PATH` checkbox during installation.'
            ])
            logger.error(msg)
            raise FileNotFoundError(msg) from exc

    @property
    def plugins_data(self) -> list[dict]:
        """Get/cache the plugin data."""
        if self.disabled:
            return []
        if self._plugin_data is None:
            data_file = resources.files('ocr_translate') / 'plugins_data.json'
            if not data_file.exists():
                self._plugin_data = []
            else:
                with open(data_file) as f:
                    self._plugin_data = json.load(f)
        return self._plugin_data

    def get_plugin_data(self, name: str) -> dict:
        """Get the data for a specific plugin."""
        for plugin in self.plugins_data:
            if plugin['name'] == name:
                return plugin
        return {}

    def initialize_scopes(self):
        """Configure the scopes."""
        if self.disabled:
            return
        for scope in self.scopes:
            pth = self.plugin_dir / scope
            pth.mkdir(exist_ok=True, parents=True)
            self.PLUGIN_SP[scope] = pth
            sys.path.insert(0, pth.as_posix())
            logger.debug(f'Added {pth} to sys.path')
        sys.path_importer_cache.clear()

        site.USER_BASE = self.plugin_dir.as_posix()
        site.USER_SITE = self.PLUGIN_SP[GENERIC_SCOPE].as_posix()

    @property
    def plugins(self) -> list[str]:
        """Get/cache the list of installed plugins."""
        if self.disabled:
            return []
        if self._plugins is None:
            self._plugins = []
            if self.plugin_list_file.exists():
                with open(self.plugin_list_file) as f:
                    self._plugins = json.load(f)
        return self._plugins

    @property
    def installed_pkgs(self) -> dict:
        """Get/cache the installed packages."""
        if self.disabled:
            return {}
        if self._installed is None:
            self._installed = {}
            if self.installed_file.exists():
                with open(self.installed_file) as f:
                    self._installed = json.load(f)

        return self._installed

    def save_plugin_list(self):
        """Save the list of installed plugins."""
        if self.disabled:
            return
        with open(self.plugin_list_file, 'w') as f:
            json.dump(self.plugins, f, indent=2)

    def save_installed(self):
        """Save the installed packages."""
        if self.disabled:
            return
        with open(self.installed_file, 'w') as f:
            json.dump(self.installed_pkgs, f, indent=2)

    @install_overrides_decorator
    def install_package(
            self,
            name: str,
            version: str,
            extras: list| str = None,
            scope: str = GENERIC_SCOPE,
            system: str = '',
            ):
        """Install a package."""
        with self.lock_pkg:
            if scope not in self.scopes:
                # logging.debug(f'Skipping {name}=={version} for scope {scope}')
                logger.debug(f'Skipping {name}=={version} for scope {scope}')
                return
            if system and system.lower() != self.system:
                logger.debug(f'Skipping {name}=={version} for system {system}')
                return

            if extras is None:
                extras = []
            elif isinstance(extras, str):
                extras = list(filter(None, extras.split(' ')))
            scoped_name = f'{scope}::{name}'

            if scoped_name in self.installed_pkgs:
                if self.installed_pkgs[scoped_name]['version'] == version:
                    logger.debug(f'{scoped_name}=={version} already installed')
                    return
                logger.debug(f'Uninstalling {scoped_name}=={self.installed_pkgs[scoped_name]["version"]}')
                self._uninstall_package(scoped_name)

            ptr = {}
            self.installed_pkgs[scoped_name] = ptr
            ptr['version'] = version
            ptr['files'] = []
            ptr['dirs'] = []

            _pip_install(f'{name}=={version}', self.plugin_dir, extras)

            tmp_dir = find_site_packages(self.plugin_dir)
            if tmp_dir is None:
                raise FileNotFoundError(f'Could not find site-packages in {self.plugin_dir}')
            logger.debug(f'Package installed in {tmp_dir}')
            for src in tmp_dir.iterdir():
                if src.name.endswith('__pycache__'):
                    continue
                # shutil.move(d, PLUGIN_SP[scope])
                dst = self.PLUGIN_SP[scope] / src.name
                if src.is_file():
                    ptr['files'].append(src.name)
                    if dst.exists():
                        dst.unlink()
                    shutil.move(src, dst)
                elif src.is_dir():
                    ptr['dirs'].append(src.name)
                    if dst.exists():
                        shutil.copytree(src, dst, dirs_exist_ok=True)
                        shutil.rmtree(src)
                    else:
                        shutil.move(src, dst)

            # Make sure only one site-packages dir exists after installing a package
            shutil.rmtree(tmp_dir)

            self.save_installed()

    def _install_plugin(self, name: str):
        """Run the `install` part of the plugin."""
        if not (data := self.get_plugin_data(name)):
            raise ValueError(f'Plugin {name} not found')
        pkg = data['package']
        version = data['version']
        deps = data.get('dependencies', [])

        logger.info(f'Installing plugin {pkg}=={version} with dependencies')
        for dep in deps[::-1]:
            self.install_package(**dep)

        self.install_package(pkg, version)

    def install_plugin(self, name: str):
        """Ensure the plugin is installed."""
        with self.lock_plugin:
            self._install_plugin(name)
            if name not in self.plugins:
                self.plugins.append(name)
                self.save_plugin_list()
            if name not in settings.INSTALLED_APPS:
                settings.INSTALLED_APPS.append(name)
            reload_django_apps()

    def _uninstall_package(self, name: str):
        """Uninstall a package."""
        if name not in self.installed_pkgs:
            return
        logger.info(f'Uninstalling package {name}')
        scope, _ = name.split('::')
        if scope not in self.scopes:
            logger.debug(f'Skipping {name} for scope {scope}')
            return
        ptr = self.installed_pkgs.pop(name)
        scope_dir = self.PLUGIN_SP[scope]
        for pth in ptr['files']:
            (scope_dir / pth).unlink()
        for pth in ptr['dirs']:
            shutil.rmtree(scope_dir / pth)

        self.save_installed()


    def uninstall_package(self, name: str):
        """Uninstall a package thread safe."""
        with self.lock_pkg:
            self._uninstall_package(name)

    def uninstall_plugin(self, name: str):
        """Uninstall the plugin."""
        with self.lock_plugin:
            if name not in self.plugins:
                return
            data = self.get_plugin_data(name)
            pkg = data['package']
            scope = data.get('scope', GENERIC_SCOPE)
            scoped_name = f'{scope}::{pkg}'
            logger.info(f'Uninstalling plugin {scoped_name} and non-shared dependencies')
            # logger.debug(f'INSTALLED: {INSTALLED}')
            if scoped_name not in self.installed_pkgs:
                raise ValueError(f'Plugin {name} not installed')

            other_deps = set()
            plugin_deps = set()
            torm_deps = set()
            for plugin in self.plugins_data:
                if plugin['name'] not in self.plugins:
                    continue
                deps = set()
                for dep in plugin.get('dependencies', []):
                    deps.add(f"{dep.get('scope', GENERIC_SCOPE)}::{dep['name']}")
                if plugin['name'] == name:
                    plugin_deps |= deps
                else:
                    other_deps |= deps
            torm_deps = plugin_deps - other_deps

            for dep in torm_deps:
                self.uninstall_package(dep)
            self.uninstall_package(scoped_name)

            self.plugins.remove(name)
            self.save_plugin_list()

            settings.INSTALLED_APPS.remove(name)
            reload_django_apps()



# Generated by Django 4.2.3 on 2024-02-01 00:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0017_alter_ocrboxmodel_iso1_map_alter_ocrmodel_iso1_map_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='ocrmodel',
            name='processor_name',
            field=models.CharField(max_length=128, null=True),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='tokenizer_name',
            field=models.CharField(max_length=128, null=True),
        ),
    ]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Initialize the server based on environment variables."""

import os
from typing import Callable

from django.db.utils import OperationalError

from .initializers import auto_create_languages, init_last_used, init_most_used

FAIL = False

def run_on_env(env_name: str, func_map: dict[str, Callable], value: str = 'true'):
    """Run a function if the environment variable is set."""
    global FAIL
    if env_name in os.environ:
        value = os.environ.get(env_name).lower()
        if value in func_map:
            try:
                func = func_map[value]
                func()
                print(f'INFO: Ran `{func.__name__}` based on environment variable `{env_name}`')
            except OperationalError as exc:
                FAIL = True
                print(f'WARNING: Ignoring environment variable `{env_name}` as the database is not ready/migrated.')
                print(f'WARNING: {exc}')
        else:
            print('Unknown value for environment variable `{env_name}`: {value}')

def deprecate_los_true():
    """Deprecate the environment variable `LOAD_ON_START=true`."""
    print('WARNING: The environment variable `LOAD_ON_START=true` is deprecated (defaults to `most`).')
    print('WARNING: Use `LOAD_ON_START=most` or `LOAD_ON_START=last` instead.')
    init_most_used()

run_on_env(
    'AUTOCREATE_LANGUAGES',
    {
        'true': auto_create_languages,
        'false': lambda: None
    }
)
run_on_env(
    'LOAD_ON_START',
    {
        'most': init_most_used,
        'last': init_last_used,
        'true': deprecate_los_true,
        'false': lambda: None
    }
)

if FAIL:
    print('WARNING: Create/migrate the database by running `python manage.py migrate`')



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
import logging
import re
from importlib.metadata import entry_points
from typing import Generator, Type, TypedDict, Union

import numpy as np
from django.db import models
from PIL.Image import Image as PILImage

from . import queues
from .messaging import Message
from .tries import get_trie_src

logger = logging.getLogger('ocr.general')

def get_or_create(model: Type[models.Model], strict: bool = False, **kwargs) -> models.Model:
    """Get or create a model instance in a safe way.

    Args:
        model (Type[models.Model]): The model class to get or create.
        strict (bool, optional): Whether to raise a caught exception if ocurred. Defaults to False.
    """
    try:
        obj, _ = model.objects.get_or_create(**kwargs)
    except model.MultipleObjectsReturned as exc:
        if strict:
            raise exc
        logger.warning(f'Multiple objects returned for {model}: {model.objects.filter(**kwargs).all()}')
        obj = model.objects.filter(**kwargs).first()
    return obj

class OptionDict(models.Model):
    """Dictionary of options for OCR and translation"""
    options = models.JSONField(unique=True)

    def __str__(self):
        return str(self.options)

class Language(models.Model):
    """Language used for translation"""
    name = models.CharField(max_length=64, unique=True)
    iso1 = models.CharField(max_length=8, unique=True)
    iso2b = models.CharField(max_length=8, unique=True)
    iso2t = models.CharField(max_length=8, unique=True)
    iso3 = models.CharField(max_length=32, unique=True)

    default_options = models.ForeignKey(
        OptionDict, on_delete=models.CASCADE, related_name='lang_default_options', null=True
        )

    load_events_src = models.ManyToManyField('LoadEvent', related_name='languages_src')
    load_events_dst = models.ManyToManyField('LoadEvent', related_name='languages_dst')

    def __str__(self):
        return f'{self.name} ({self.iso1})'

    def __eq__(self, other):
        if isinstance(other, Language):
            return self.iso1 == other.iso1
        if isinstance(other, str):
            return self.iso1 == other
        return False

    # https://stackoverflow.com/questions/61212514/django-model-objects-became-not-hashable-after-upgrading-to-django-2-2
    def __hash__(self):
        return hash(self.iso1)

    def load_src(self) -> None:
        """Load the language trie"""
        self.load_events_src.create(description=f'Loading SRC Language {self.name}')

    def load_dst(self) -> None:
        """Load the language trie"""
        self.load_events_dst.create(description=f'Loading DST Language {self.name}')

    @classmethod
    def get_last_loaded_src(cls) -> 'Language':
        """Get the last loaded language"""
        res = cls.objects
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events_src'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events_src__date')
        res = res.first()
        if res is None or len(res.load_events_src.all()) == 0:
            logger.debug('No load events found for Language')
            return None
        return res

    @classmethod
    def get_last_loaded_dst(cls) -> 'Language':
        """Get the last loaded language"""
        res = cls.objects
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events_dst'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events_dst__date')
        res = res.first()
        if res is None or len(res.load_events_dst.all()) == 0:
            logger.debug('No load events found for Language')
            return None
        return res

class LoadEvent(models.Model):
    """Event log for the OCR and translation tasks"""
    description = models.CharField(max_length=512)

    date = models.DateTimeField(auto_now_add=True)

class BaseModel(models.Model):
    """Mixin class for loading entrypoint models"""
    class Meta:
        abstract = True
    # This should be a dict of dicts where the key is the name of the option and the value contains:
    #  - type: The type of the option (str, int, float, bool)
    #  - default: The default value of the option (can be a callable that returns the default value)
    #  - description: A description of the option
    ALLOWED_OPTIONS = {}

    entrypoint_namespace = None

    name = models.CharField(max_length=128)

    entrypoint = models.CharField(max_length=128, null=True)

    language_format = models.CharField(max_length=32, null=True)
    iso1_map = models.JSONField(null=True, blank=True)

    active = models.BooleanField(default=False)

    default_options = models.ForeignKey(
        OptionDict, on_delete=models.SET_NULL, related_name='used_by_%(class)s', null=True
        )

    load_events = models.ManyToManyField(LoadEvent, related_name='%(class)s')

    def __str__(self):
        return str(self.name)

    def __del__(self):
        try:
            self.unload()
        except NotImplementedError:
            pass

    def __getattribute__(self, name):
        res = super().__getattribute__(name)
        if name == 'load':
            def wrapped():
                _res = res()
                self.load_events.create(description=f'Loading model {self.name}')
                return _res
            return wrapped
        return res

    @classmethod
    def get_last_loaded(cls) -> 'BaseModel':
        """Get the last loaded model"""
        res = cls.objects
        res = res.filter(active=True)
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events__date')
        res = res.first()
        if res is None or len(res.load_events.all()) == 0:
            logger.debug(f'No load events found for {cls.__name__}')
            return None
        return res

    def get_lang_code(self, lang: 'Language') -> str:
        """Get the language code for a specific model"""
        if isinstance(self.iso1_map, dict) and lang.iso1 in self.iso1_map:
            return self.iso1_map[lang.iso1]
        return getattr(lang, self.language_format or 'iso1')

    @classmethod
    def from_entrypoint(cls, name: str) -> Type['models.Model']:
        """Get the entrypoint specific TSL model class from the entrypoint name"""
        if cls.entrypoint_namespace is None:
            raise ValueError('Cannot load base model class from entrypoint.')

        obj = cls.objects.get(name=name)
        ept = obj.entrypoint

        logger.debug(f'Loading model {name} from entrypoint {cls.entrypoint_namespace}:{ept}')
        for entrypoint in entry_points(group=cls.entrypoint_namespace, name=ept):
            new_cls = entrypoint.load()
            break
        else:
            raise ValueError(f'Missing plugin: Entrypoint "{ept}" not found.')

        return new_cls.objects.get(name=name)

    def load(self) -> None:
        """Placeholder method for loading the model. To be implemented via entrypoint"""
        raise NotImplementedError('The base model class does not implement this method.')

    def unload(self) -> None:
        """Placeholder method for unloading the model. To be implemented via entrypoint"""
        raise NotImplementedError('The base model class does not implement this method.')

class OCRModel(BaseModel):
    """OCR model."""
    entrypoint_namespace = 'ocr_translate.ocr_models'
    # iso1 code for languages that do not use spaces to separate words
    _NO_SPACE_LANGUAGES = ['ja', 'zh', 'zht', 'lo', 'my']
    _VERTICAL_LANGS = ['ja', 'zh', 'zht', 'ko']
    SINGLE='single'
    MERGED='merged'
    OCR_MODE_CHOICES = [
        (SINGLE, 'Single'),
        (MERGED, 'Merged'),
    ]

    languages = models.ManyToManyField(Language, related_name='ocr_models')
    ocr_mode = models.CharField(max_length=32, choices=OCR_MODE_CHOICES, default=MERGED)
    tokenizer_name = models.CharField(max_length=128, null=True)
    processor_name = models.CharField(max_length=128, null=True)

    def prepare_image(
            self,
            img: PILImage, bbox: tuple[int, int, int, int] = None
            ) -> PILImage:
        """Standard operation to be performed on image before OCR. E.G color scale and crop to bbox"""
        if not isinstance(img, PILImage):
            raise TypeError(f'img should be PIL Image, but got {type(img)}')
        img = img.convert('RGB')

        if bbox:
            img = img.crop(bbox)

        return img

    @staticmethod
    def merge_single_result( # pylint: disable=too-many-locals
            lang: str,
            texts: list[str],
            bboxes_single: list['BBox'],
            bboxes_merged: list['BBox'],
            ) -> list[str]:
        """Merge the results of an OCR run on the single components of a detected bounding box.
        Will try to understand if the text is horizontal or vertical and merge accordingly.

        Args:
            lang (str): Language of the text in iso1 format.
            texts (list[str]): List of texts for each component.
            bboxes_single (list['BBox']): List of single bounding boxes.
            bboxes_merged (list['BBox']): List of merged bounding boxes.

        Returns:
            list[str]: The merged text.
        """
        grouped = {}
        # Group texts and single_bboxes by merged bounding box
        for text, bbox in zip(texts, bboxes_single):
            merged = bbox.to_merged
            ptr = grouped.setdefault(merged, [])
            ptr.append((text, *bbox.lbrt))

        sep = '' if lang in OCRModel._NO_SPACE_LANGUAGES else ' '

        res = []
        for bbox in bboxes_merged:
            single_array = np.array(grouped[bbox], dtype=object)
            l,b,r,t = bbox.lbrt
            width = r - l
            height = t - b

            # Vertical langs can also be written horizontal
            # Assume that if box.width > box.height * 1.3 then it's horizontal
            vertical = lang in OCRModel._VERTICAL_LANGS and height * 1.3 > width

            if vertical:
                # Vertical text find columns
                thr = np.average(single_array[:, 3] - single_array[:, 1]) / 1.5
                centers = (single_array[:, 1] + single_array[:, 3]) / 2
            else:
                # Horizontal text find lines
                thr = np.average(single_array[:, 4] - single_array[:, 2]) / 1.5
                centers = (single_array[:, 2] + single_array[:, 4]) / 2

            # Group lines/cols if the centers are closer than the average line width/height (vertical/non-vert)
            classifiers = np.empty(0)
            line_classifiers = []
            for cen in centers:
                if len(classifiers) == 0:
                    classifiers = np.array([cen])
                    line_classifiers.append(0)
                    continue
                # Index of classifier closest to the center
                i_min = np.argmin(np.abs(classifiers - cen))
                # If the closest classifier is closer than the threshold, reuse it
                if np.abs(classifiers[i_min] - cen) < thr:
                    line_classifiers.append(i_min)
                # Otherwise create a new classifier
                else:
                    classifiers = np.append(classifiers, cen)
                    line_classifiers.append(len(classifiers) - 1)

            line_classifiers = np.array(line_classifiers)
            # Sort lines/cols top-to-bottom or right-to-left
            i_sort = np.argsort(classifiers)[::-1 if vertical else 1]
            text_chunks = []
            # Sort lines/cols chunks left-to-right or top-to-bottom
            sort_col_index = 4 if vertical else 1
            for i in i_sort:
                ind = np.where(line_classifiers == i)[0]
                j_sort = np.argsort(single_array[ind, sort_col_index])
                text_chunks += single_array[ind[j_sort], 0].tolist()

            res.append(sep.join(text_chunks))

        return res

    def _ocr(
            self,
            img: PILImage, lang: str = None, options: dict = None
            ) -> str:
        """PLACEHOLDER (to be implemented via entrypoint): Perform OCR on an image.

        Args:
            img (Image.Image):  A Pillow image on which to perform OCR.
            lang (str, optional): The language to use for OCR. (Not every model will use this)
            bbox (tuple[int, int, int, int], optional): The bounding box of the text on the image in lbrt format.
            options (dict, optional): A dictionary of options to pass to the OCR model.

        Raises:
            TypeError: If img is not a Pillow image.

        Returns:
            str: The text extracted from the image.
        """
        # Redefine this method with the same signature as above
        # Should return a string with the result of the OCR performed on the input PILImage.
        # Unless the methods `prepare_image` or `ocr` are also being overwritten, the input image will be the
        #  result of the CROP on the original image using the bounding boxes given by the box detection model.
        raise NotImplementedError('The base model class does not implement this method.')

    def ocr(
            self,
            bbox_obj: 'BBox', lang: 'Language',  image: PILImage = None, options: 'OptionDict' = None,
            force: bool = False, block: bool = True,
            ) -> Generator[Union[Message, 'Text'], None, None]:
        """High level function to perform OCR on an image.

        Args:
            bbox_obj (m.BBox): The BBox object from the database.
            lang (m.Language): The Language object from the database.
            image (Image.Image, optional): The image on which to perform OCR. Needed if no previous OCR run exists, or
                force is True.
            options (m.OptionDict, optional): The OptionDict object from the database
                containing the options for the OCR.
            force (bool, optional): Whether to force the OCR to run again even if a previous run exists.
                Defaults to False.
            block (bool, optional): Whether to block until the task is complete. Defaults to True.

        Raises:
            ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

        Yields:
            Generator[Union[Message, m.Text], None, None]:
                If block is False, yields a Message object for the OCR run first and the resulting Text object second.
                If block is True, yields the resulting Text object.
        """
        options_obj = options
        if options_obj is None:
            options_obj = OptionDict.objects.get(options={})
        params = {
            'bbox': bbox_obj,
            'model': self,
            'lang_src': lang,
            'options': options_obj,
        }
        ocr_run_obj = OCRRun.objects.filter(**params).first()
        if ocr_run_obj is None or force:
            if image is None:
                raise ValueError('Image is required for OCR')
            logger.info('Running OCR')

            id_ = (bbox_obj.id, self.id, lang.id, options_obj.id)
            mlang = self.get_lang_code(lang)
            opt_dct = options_obj.options
            text = queues.ocr_queue.put(
                id_=id_,
                handler=self._ocr,
                msg={
                    'args': (self.prepare_image(image, bbox_obj.lbrt),),
                    'kwargs': {
                        'lang': mlang,
                        'options': opt_dct
                        },
                },
            )
            if not block:
                yield text
            text = text.response()
            if lang.iso1 in self._NO_SPACE_LANGUAGES:
                text = text.replace(' ', '')

            text_obj = get_or_create(Text, text=text)
            params[f'result_{self.ocr_mode}'] = text_obj
            ocr_run_obj = OCRRun.objects.create(**params)
        else:
            if not block:
                # Both branches should have the same number of yields
                yield None
            logger.info(f'Reusing OCR <{ocr_run_obj.id}>')
            # It is possible that single has to be performed is a previous pipeline was interrupted
            text_obj = ocr_run_obj.result_merged or ocr_run_obj.result_single
            # text = ocr_run.result.text

        yield text_obj


class BoxDetectionResult(TypedDict):
    """Type for the result of the box detection"""
    single: list[tuple[int, int, int, int]]
    merged: tuple[int, int, int, int]

class OCRBoxModel(BaseModel):
    """OCR model for bounding boxes"""
    #pylint: disable=abstract-method
    entrypoint_namespace = 'ocr_translate.box_models'

    languages = models.ManyToManyField(Language, related_name='box_models')

    def _box_detection(
            self,
            image: PILImage, options: dict = None
            ) -> list[BoxDetectionResult]:
        """PLACEHOLDER (to be implemented via entrypoint): Perform box OCR on an image.
        Returns list of bounding boxes as dicts:
            - merged: The merged BBox as a tuple[int, int, int, int]
            - single: List of BBoxed merged into the merged BBox as a tuple[int, int, int, int]

        Args:
            image (Image.Image): A Pillow image on which to perform OCR.
            options (dict, optional): A dictionary of options.

        Raises:
            NotImplementedError: The type of model specified is not implemented.

        Returns:
            list[BoxDetectionResult]: List of dictionary with key/value pairs:
              - merged: The merged BBox as a tuple[int, int, int, int]
              - single: List of BBoxed merged into the merged BBox as a tuple[int, int, int, int]
        """
        # Redefine this method with the same signature as above
        # Should return a list of `lrbt` boxes after processing the input PILImage
        raise NotImplementedError('The base model class does not implement this method.')

    def box_detection( # pylint: disable=too-many-locals
            self,
            img_obj: 'Image', lang: 'Language', image: PILImage = None,
            force: bool = False, options: 'OptionDict' = None
            ) -> tuple[list['BBox'], list['BBox']]:
        """High level function to perform box OCR on an image. Will attempt to reuse a previous run if possible.

        Args:
            img_obj (m.Image): An Image object from the database.
            lang (m.Language): A Language object from the database (not every model is gonna use this).
            image (Image.Image, optional): The Pillow image to be used for OCR if a previous result is not found.
                Defaults to None.
            force (bool, optional): If true, re-run the OCR even if a previous result is found. Defaults to False.
            options (m.OptionDict, optional): An OptionDict object from the database. Defaults to None.

        Raises:
            ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

        Returns:
            tuple[list['BBox'], list['BBox']]: Tuple of lists of BBox objects from the database.
                First list is the single bounding boxes, second list is the merged bounding boxes.
        """
        options_obj = options or OptionDict.objects.get(options={})
        params = {
            'image': img_obj,
            'model': self,
            'options': options_obj,
            'lang_src': lang,
        }

        bbox_run = OCRBoxRun.objects.filter(**params).first()
        # Needed to rerun the OCR from <0.4.x to >=0.4.x
        # Before only merged boxes where saved, now also the single are needed
        if isinstance(bbox_run, OCRBoxRun):
            if len(bbox_run.result_single.all()) == 0 and len(bbox_run.result_merged.all()) > 0:
                bbox_run.delete()
                bbox_run = None
        if bbox_run is None or force:
            if image is None:
                raise ValueError('Image is required for BBox OCR')
            logger.info('Running BBox OCR')
            opt_dct = options_obj.options
            id_ = (img_obj.id, self.id, lang.id, options_obj.id)
            bboxes = queues.box_queue.put(
                id_=id_,
                handler=self._box_detection,
                msg={
                    'args': (image,),
                    'kwargs': {'options': opt_dct},
                },
            )
            # bboxes_single, bboxes_merged = bboxes.response()
            bboxes_list = bboxes.response()
            # Create it here to avoid having a failed entry in DB
            bbox_run = OCRBoxRun.objects.create(**params)
            for dct in bboxes_list:
                merged = dct['merged']
                l,b,r,t = merged
                bbox_merged_obj = BBox.objects.create(
                    l=l, b=b, r=r, t=t,
                    image=img_obj,
                    from_ocr_merged=bbox_run,
                    )
                for bbox in dct['single']:
                    l,b,r,t = bbox
                    BBox.objects.create(
                        l=l, b=b, r=r, t=t,
                        image=img_obj,
                        from_ocr_single=bbox_run,
                        to_merged=bbox_merged_obj,
                        )
        else:
            logger.info(f'Reusing BBox OCR <{bbox_run.id}>')

        res_single = list(bbox_run.result_single.all())
        res_merged = list(bbox_run.result_merged.all())
        logger.debug(f'BBox OCR result: {len(res_single)} single boxes')
        logger.info(f'BBox OCR result: {len(res_merged)} merged boxes')

        return res_single, res_merged


class TSLModel(BaseModel):
    """Translation models using hugging space naming convention"""
    ALLOWED_OPTIONS = {
        'ignore_chars': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': 'Characters to ignore during translation.',
        },
        'break_chars': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': (
                'Characters on which lines will be split for translation.\n'
                'The split lines are translated separately (no context between them) and then merged back together.)'
                ),
        },
        'allowed_start_end': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': (
                'Characters allowed at the start and end of a line.\n'
                'Some models like tesseract, will detect the edges of text bubbles as text and spit out garbage '
                'at the start/end of the translated line. This option will filter out any caracter that is not '
                'allowed at the start/end of a line.'
                ),
        },
        'break_newlines': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': 'Split lines on newlines. If false the newlines will be replaced with spaces.',
        },
        'restore_missing_spaces': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': (
                'Some models will OCR text without spaces between some of the words.\n'
                'This option will attempt to restore the missing spaces by finding the shortest valid decomposition '
                'also keeping into account the word frequency.\n'
                '(Only available if a trie is loaded for the source language.)'
                ),
        },
        'restore_dash_newlines': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': (
                'Text can be written with dash splitting a word onto a new line.\n'
                'If this option is enabled, dashes preceded by a character and followed by a newline will be removed '
                'togheter with the newline to merge the word back together.'
                ),
        },
    }
    entrypoint_namespace = 'ocr_translate.tsl_models'

    src_languages = models.ManyToManyField(Language, related_name='tsl_models_src')
    dst_languages = models.ManyToManyField(Language, related_name='tsl_models_dst')

    @staticmethod
    def pre_tokenize( # pylint: disable=too-many-branches
            text: str,
            ignore_chars: str = None,
            break_chars: str = None,
            allowed_start_end: str = None,
            break_newlines: bool = False,
            restore_missing_spaces: bool = False,
            restore_dash_newlines: bool = False,
            **kwargs
            ) -> list[str]:
        """Pre-tokenize a text string.

        Args:
            text (str): Text to tokenize.
            lang (str): Language of the text.
            ignore_chars (str, optional): String of characters to ignore. Defaults to None.
            break_chars (str, optional): String of characters to break on. Defaults to None.
            allowed_start_end (str, optional): String of characters allowed at the start and end of a line.
            break_newlines (bool, optional): Whether to break on newlines. Defaults to True.
            restore_missing_spaces (bool, optional): Whether to restore missing spaces (2 word with no space between).
            restore_dash_newlines (bool, optional): Whether to restore dash-newlines (word broken with a -newline).
                Defaults to False.

        Returns:
            list[str]: List of string tokens.
        """
        if isinstance(break_newlines, str):
            break_newlines = break_newlines.lower() == 'true'
        if isinstance(restore_missing_spaces, str):
            restore_missing_spaces = restore_missing_spaces.lower() == 'true'
        if isinstance(restore_dash_newlines, str):
            restore_dash_newlines = restore_dash_newlines.lower() == 'true'
        orig_text = text
        if allowed_start_end is not None:
            rgx_start = re.compile(
                '(?x)'
                rf'^[^{allowed_start_end}]+\S?(?= )'
                '|'
                rf'^\S[^{allowed_start_end}]+(?= )'
                )

            rgx_end = re.compile(
                '(?x)'
                rf'(?<= )\S?[^{allowed_start_end}]+$'
                '|'
                rf'(?<= )[^{allowed_start_end}]+\S$'
                )

            app = []
            for split in text.split('\n'):
                split = rgx_start.sub('', split)
                split = rgx_end.sub('', split)
                app.append(split)
            text = '\n'.join(app)
        if restore_dash_newlines:
            text = re.sub(r'(?<!\n)- *\n', '', text)
        if ignore_chars:
            text = re.sub(f'[{ignore_chars}]+', '', text)
        if break_chars is None:
            break_chars = ''
        if break_newlines:
            break_chars += '\n'
        else:
            text = text.replace('\n', ' ')

        if restore_missing_spaces and not (trie := get_trie_src()) is None:
            res = [
                trie.decompose(split, min_length=1)
                if not trie.search(split, strict=False) else
                [[split]]
                for split in text.lower().split(' ')
                ]

            # Use a list of word frequencies to determine the best split
            def sum_freq(lst: list) -> float:
                return sum(trie.get_freq(w) for w in lst) / len(lst)**4.0

            res = [' '.join(max(_, key=sum_freq)) for _ in filter(None, res)]
            text = ' '.join(res)

        break_chars = re.escape(break_chars)
        tokens = text
        if len(break_chars) > 0:
            tokens = re.split(f'[{break_chars}+]', text)

        if isinstance(tokens, str):
            tokens = [text]

        res = list(filter(None, tokens))
        logger.debug(f'Pre-tokenized "{orig_text}" to {res}')
        return res if len(res) > 0 else [' ']


    def _translate(
            self,
            tokens: list, src_lang: str, dst_lang: str, options: dict = None) -> str | list[str]:
        """PLACEHOLDER (to be implemented via entrypoint): Translate a text using a the loaded model.

        Args:
            tokens (list): list or list[list] of string tokens to be translated.
            lang_src (str): Source language.
            lang_dst (str): Destination language.
            options (dict, optional): Options for the translation. Defaults to {}.

        Raises:
            TypeError: If text is not a string or a list of strings.

        Returns:
            Union[str,list[str]]: Translated text. If text is a list, returns a list of translated strings.
        """
        # Redefine this method with the same signature as above
        # Should return a string with the translated text.
        # IMPORTANT: the main codebase treats this function as batchable:
        # The input `tokens` can be a list of strings or a list of list of strings. The output should match the input
        #   being a string or list of strings.
        # (This is used to leverage the capability of pytorch to batch inputs and outputs for faster performances,
        #   or it can also used to write a plugin for an online service by using a single request for multiple inputs
        #   using some separator that the service will leave unaltered.)
        raise NotImplementedError('The base model class does not implement this method.')

    def find_manual(self,  text_obj: 'Text', src: 'Language', dst: 'Language') -> 'TranslationRun':
        """Find a manual translation run for the given text with the specified source and destination languages.
        Args:
            text_obj (m.Text): Text object from the database to translate.
            src (m.Language): Source language object from the database.
            dst (m.Language): Destination language object from the database.

        Returns:
            m.TranslationRun: The TranslationRun object from the database.
        """
        manual_model, _ = TSLModel.objects.get_or_create(name='manual')
        params = {
            'model': manual_model,
            'lang_src': src,
            'lang_dst': dst,
            'text': text_obj,
            'options': OptionDict.objects.get(options={})
        }
        # logger.debug(f'Looking for manual TSL with params: {params}')
        return TranslationRun.objects.filter(**params).first()


    def translate(
            self,
            text_obj: 'Text', src: 'Language', dst: 'Language', options: 'OptionDict' = None,
            force: bool = False,
            block: bool = True,
            favor_manual: bool = True,
            lazy: bool = False
            ) -> Generator[Union[Message, 'Text'], None, None]:
        """High level translate call generating a TranslationRun entry.
        Args:
            text_obj (m.Text): Text object from the database to translate.
            src (m.Language): Source language object from the database.
            dst (m.Language): Destination language object from the database.
            options (m.OptionDict, optional): OptionDict object from the database. Defaults to None.
            force (bool, optional): Whether to force a new TSL run. Defaults to False.
            block (bool, optional): Whether to block until the task is complete. Defaults to True.
            favor_manual (bool, optional): Whether to favor manual translations over TSL. Defaults to True.
            lazy (bool, optional): Whether to raise an error if the TSL run is not found. Defaults to False.

        Raises:
            ValueError: If lazy and force are both True or if lazy is True and the TSL run is not found.

        Yields:
            Generator[Union[Message, m.Text], None, None]:
                If block is False, yields a Message object for the TSL run first and the resulting Text object second.
                If block is True, yields the resulting Text object.
        """
        if lazy and force:
            raise ValueError('Cannot force + lazy TSL run')
        # Check if a ManualModel is being used
        tsl_run_obj = None
        if favor_manual:
            tsl_run_obj = self.find_manual(text_obj, src, dst)
        if tsl_run_obj is None:
            options_obj = options or OptionDict.objects.get(options={})
            params = {
                'options': options_obj,
                'text': text_obj,
                'model': self,
                'lang_src': src,
                'lang_dst': dst,
            }
            tsl_run_obj = TranslationRun.objects.filter(**params).first()
        if tsl_run_obj is None or force:
            if lazy:
                raise ValueError('Value not found for lazy TSL run')
            logger.info('Running TSL')
            # Generate a unique id for a message
            id_ = (text_obj.id, self.id, options_obj.id, src.id, dst.id, options_obj.id)
            batch_id = (self.id, options_obj.id, src.id, dst.id, options_obj.id)
            lang_dct = getattr(src.default_options, 'options', {})
            model_dct =  getattr(self.default_options, 'options', {})
            opt_dct = {**lang_dct, **model_dct, **options_obj.options}

            tokens = self.pre_tokenize(text_obj.text, **opt_dct)
            new = queues.tsl_queue.put(
                id_=id_,
                batch_id=batch_id,
                handler=self._translate,
                msg={
                    'args': (
                        tokens,
                        self.get_lang_code(src),
                        self.get_lang_code(dst),
                        ),
                    'kwargs': {'options': opt_dct},
                },
            )
            if not block:
                yield new
            new = new.response()
            text_obj = get_or_create(Text, text=new)
            params['result'] = text_obj
            tsl_run_obj = TranslationRun.objects.create(**params)
        else:
            if not block:
                # Both branches should have the same number of yields
                yield None
            manual = ''
            if tsl_run_obj.model.name == 'manual':
                manual = 'manual '
            logger.info(f'Reusing {manual}TSL <{tsl_run_obj.id}>')

        yield tsl_run_obj.result

class Image(models.Model):
    """Image registered as the md5 of the uploaded file"""
    md5 = models.CharField(max_length=32, unique=True)

class BBox(models.Model):
    """Bounding box of a text in an image"""
    l = models.IntegerField(null=False)
    b = models.IntegerField(null=False)
    r = models.IntegerField(null=False)
    t = models.IntegerField(null=False)

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='bboxes')
    from_ocr_single = models.ForeignKey(
        'OCRBoxRun', on_delete=models.CASCADE, related_name='result_single',
        default=None, null=True
        )
    from_ocr_merged = models.ForeignKey(
        'OCRBoxRun', on_delete=models.CASCADE, related_name='result_merged',
        default=None, null=True
        )
    to_merged = models.ForeignKey(
        'BBox', on_delete=models.CASCADE, related_name='from_single',
        default=None, null=True
        )

    @property
    def lbrt(self):
        """Return the bounding box as a tuple of (left, bottom, right, top)"""
        return self.l, self.b, self.r, self.t

    def __str__(self):
        return f'{self.lbrt}'

class Text(models.Model):
    """Text extracted from an image or translated from another text"""
    text = models.TextField()
    # lang = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='texts')

class OCRBoxRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_box_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='box_src')

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_box')
    model = models.ForeignKey(OCRBoxModel, on_delete=models.CASCADE, related_name='box_runs')
    # result = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='from_ocr')

class OCRRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='ocr_src')

    # image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_ocr')
    bbox = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='to_ocr')
    model = models.ForeignKey(OCRModel, on_delete=models.CASCADE, related_name='ocr_runs')
    result_single = models.ForeignKey(
        Text, on_delete=models.CASCADE, related_name='from_ocr_single',
        default=None, null=True
        )
    result_merged = models.ForeignKey(
        Text, on_delete=models.CASCADE, related_name='from_ocr_merged',
        default=None, null=True
        )

class TranslationRun(models.Model):
    """Translation run on a text using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='trans_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_src')
    lang_dst = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_dst')

    text = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='to_trans')
    model = models.ForeignKey(TSLModel, on_delete=models.CASCADE, related_name='tsl_runs')
    result = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='from_trans')



# Generated by Django 4.2.4 on 2023-09-20 11:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0008_ocrboxmodel_entrypoint_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='language',
            name='easyocr',
        ),
        migrations.RemoveField(
            model_name='language',
            name='facebookM2M',
        ),
        migrations.RemoveField(
            model_name='language',
            name='tesseract',
        ),
        migrations.AddField(
            model_name='ocrboxmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='tslmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
    ]



# Generated by Django 5.1 on 2025-02-10 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0019_ocrboxmodel_active_ocrmodel_active_tslmodel_active'),
    ]

    operations = [
        migrations.CreateModel(
            name='LoadEvent',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('description', models.CharField(max_length=512)),
                ('date', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='language',
            name='load_events_dst',
            field=models.ManyToManyField(
                related_name='languages_dst', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='language',
            name='load_events_src',
            field=models.ManyToManyField(
                related_name='languages_src', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='ocrboxmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='tslmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
    ]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Decorators to preparse requests and responses."""

import json
import time
from functools import wraps
from threading import Lock

from django.http import HttpRequest, JsonResponse

from .ocr_tsl.box import get_box_model
from .ocr_tsl.lang import get_lang_dst, get_lang_src
from .ocr_tsl.ocr import get_ocr_model
from .ocr_tsl.tsl import get_tsl_model

locks = {}

def use_lock(lock_name: str, blocking: bool = False):
    """Decorator to use a lock for the function."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lock = locks.setdefault(lock_name, Lock())
            if blocking:
                with lock:
                    return func(*args, **kwargs)
            else:
                acquired = lock.acquire(blocking=False)
                try:
                    return func(*args, **kwargs)
                finally:
                    if acquired:
                        lock.release()
        return wrapper
    return decorator

def wait_for_lock(lock_name: str, timeout: int = None):
    """Decorator to wait for a lock before executing the function without acquiring the lock."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lock = locks.get(lock_name, Lock())
            start = time.time()
            while lock.locked():
                time.sleep(0.1)
                if timeout is not None and time.time() - start > timeout:
                    return JsonResponse({'error': 'Timeout waiting for lock'}, status=408)

            return func(*args, **kwargs)
        return wrapper
    return decorator

def get_backend_models(strict: bool = True):
    """Decorator to check and add loaded models."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            box_model = get_box_model() or ''
            ocr_model = get_ocr_model() or ''
            tsl_model = get_tsl_model() or ''
            if strict and (not box_model or not ocr_model or not tsl_model):
                return JsonResponse({'error': 'Models not loaded'}, status=513)
            return func(*args, **kwargs, box_model=box_model, ocr_model=ocr_model, tsl_model=tsl_model)
        return wrapper
    return decorator

def get_backend_langs(strict: bool = True):
    """Decorator to check and add loaded languages."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lang_src = get_lang_src() or ''
            lang_dst = get_lang_dst() or ''
            if strict and (not lang_src or not lang_dst):
                return JsonResponse({'error': 'Languages not loaded'}, status=512)
            return func(*args, **kwargs, lang_src=lang_src, lang_dst=lang_dst)
        return wrapper
    return decorator

def method_or_405(methods):
    """Decorator to allow only specific methods."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            request: HttpRequest = args[0]
            if request.method not in methods:
                return JsonResponse({'error': f'{request.method} not allowed'}, status=405)
            return func(*args, **kwargs)
        return wrapper
    return decorator

def post_data_deserializer(expected_keys: list[str], strict: bool = True, required: bool = True):
    """Decorator to deserialize POST data.

    Args:
        expected_keys (list[str]): List of keys to expect in the POST data.
        strict (bool, optional): If any key not present in `expected_keys` is found, return 400. Defaults to True.
        required (bool, optional): If `True` not found expected keys are set to None, else return 400. Defaults to True.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            request: HttpRequest = args[0]
            if request.content_type != 'application/json':
                return JsonResponse({'error': 'Content-Type must be application/json'}, status=400)
            _data = json.loads(request.body)
            data = {}
            for key in expected_keys:
                if key not in _data and required:
                    return JsonResponse({'error': f'{key} not found in POST data'}, status=400)
                data[key] = _data.pop(key, None)
            if strict and _data:
                return JsonResponse({'error': f'Unexpected keys: {", ".join(_data.keys())}'}, status=400)
            return func(*args, **kwargs, **data)
        return wrapper
    return decorator

def get_data_deserializer(expected_keys: list[str], strict: bool = True, required: bool = True):
    """Decorator to deserialize GET data.

    Args:
        expected_keys (list[str]): List of keys to expect in the GET data.
        strict (bool, optional): If any key not present in `expected_keys` is found, return 400. Defaults to True.
        required (bool, optional): If `True` not found expected keys are set to None, else return 400. Defaults to True.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            request: HttpRequest = args[0]
            _data = request.GET.dict()
            data = {}
            for key in expected_keys:
                if key not in _data and required:
                    return JsonResponse({'error': f'{key} not found in GET data'}, status=400)
                data[key] = _data.pop(key, None)
            if strict and _data:
                return JsonResponse({'error': f'Unexpected keys: {", ".join(_data.keys())}'}, status=400)
            return func(*args, **kwargs, **data)
        return wrapper
    return decorator



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Admin interface for ocr_translate app."""
from importlib.metadata import entry_points

from django import forms
from django.contrib import admin

from . import models as m


class AllowEmptyJSONField(forms.JSONField):
    """JSONField that allows empty values"""
    empty_values = [None, '', [], ()]

class OptionDictForm(forms.ModelForm):
    """Form for OptionDict model"""
    options = AllowEmptyJSONField(required=False)

@admin.register(m.OptionDict)
class OptionDictAdmin(admin.ModelAdmin):
    """Admin interface for OptionDict model"""
    list_display = ('options',)
    form = OptionDictForm

@admin.register(m.Language)
class LanguageAdmin(admin.ModelAdmin):
    """Admin interface for Language model"""
    list_display = ('name', 'iso1', 'iso2b', 'iso2t', 'iso3')

class OCRBoxModelForm(forms.ModelForm):
    """Form for OCRBoxModel model"""
    entrypoint = forms.ChoiceField(
        choices=[(ep.name, ep.name) for ep in entry_points(group=m.OCRBoxModel.entrypoint_namespace)]
        )
    iso1_map = AllowEmptyJSONField(required=False)

class OCRModelForm(forms.ModelForm):
    """Form for OCRModel model"""
    entrypoint = forms.ChoiceField(
        choices=[(ep.name, ep.name) for ep in entry_points(group=m.OCRModel.entrypoint_namespace)]
        )
    iso1_map = AllowEmptyJSONField(required=False)

class TSLModelForm(forms.ModelForm):
    """Form for TSLModel model"""
    entrypoint = forms.ChoiceField(
        choices=[(ep.name, ep.name) for ep in entry_points(group=m.TSLModel.entrypoint_namespace)]
        )
    iso1_map = AllowEmptyJSONField(required=False)

@admin.register(m.OCRBoxModel)
class OCRBoxModelAdmin(admin.ModelAdmin):
    """Admin interface for OCRBoxModel model"""
    list_display = ('name',)
    filter_horizontal = ('languages',)
    form = OCRBoxModelForm

@admin.register(m.OCRModel)
class OCRModelAdmin(admin.ModelAdmin):
    """Admin interface for OCRModel model"""
    list_display = ('name',)
    filter_horizontal = ('languages',)
    form = OCRModelForm

@admin.register(m.TSLModel)
class TSLModelAdmin(admin.ModelAdmin):
    """Admin interface for TSLModel model"""
    list_display = ('name',)
    filter_horizontal = ('src_languages', 'dst_languages')
    form = TSLModelForm



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Trie data structure implementation."""

class TrieNode:
    """Node of the Trie data structure."""
    def __init__(self):
        self.children = {}
        self.is_word = False
        self.freq = 0.0

    def __str__(self) -> str:
        return f'TrieNode(childs: {list(self.children.keys())}, is_word: {self.is_word})'

    def __contains__(self, key: str) -> bool:
        return key in self.children

    def __getitem__(self, key: str) -> 'TrieNode':
        return self.children.get(key, None)

    def __setitem__(self, key: str, value: 'TrieNode') -> None:
        self.children[key] = value

class Trie:
    """Trie data structure."""
    def __init__(self):
        self.root = TrieNode()
        self.charset = set()

    def insert(self, word: str, freq: float = 0.0) -> None:
        """Insert a word in the Trie."""
        self.charset.update(word)

        node = self.root

        for char in word:
            node = node.children.setdefault(char, TrieNode())

        node.is_word = True
        node.freq = freq

    def search(self, word: str, strict: bool = False, node = None) -> bool:
        """Search a word in the Trie.

        Args:
            word (str): Word to search.
            strict (bool, optional): If True, return False if the word contains characters not in the Trie.
                If False, ignore characters not in the Trie.
                Defaults to False.
            node (TrieNode, optional): Current node in the recursion. Defaults to None.

        Returns:
            bool: True if the word is in the Trie, False otherwise.
        """
        if node is None:
            node = self.root

        for char in word:
            if char not in self.charset:
                if strict:
                    return False
                continue
            if char not in node.children:
                return False

            node = node.children[char]

        if node is self.root and not strict:
            return True

        return node.is_word

    def get_freq(self, word: str) -> float:
        """Return the frequency of a word in the Trie.

        Args:
            word (str): Word to search.

        Returns:
            float: Frequency of the word in the Trie.
        """
        node = self.root

        for char in word:
            if char not in node.children:
                return 0.0

            node = node.children[char]

        if not node.is_word:
            return 0.0

        return node.freq

    def autocomplete(self, prefix: str) -> list[str]:
        """Autocomplete a prefix.

        Args:
            prefix (str): Prefix to autocomplete.

        Returns:
            list[str]: List of words that start with the prefix.
        """
        node = self.root

        for char in prefix:
            if char not in node.children:
                return []

            node = node.children[char]

        res = self._autocomplete(node, prefix)

        res.sort(key=self.get_freq, reverse=True)

        return res

    def _autocomplete(self, node: TrieNode, prefix: str) -> list[str]:
        """Internal function to autocomplete a prefix.

        Args:
            node (TrieNode): Current node in the recursion.
            prefix (str): Current prefix in the recursion.

        Returns:
            list[str]: List of words that start with the prefix.
        """
        results = []

        if node.is_word:
            results.append(prefix)

        for char, child in node.children.items():
            results.extend(self._autocomplete(child, prefix + char))

        return results

    def decompose(self, word: str, min_length: int = 3) -> list[list[str]]:
        """Fing all combinations of existing words that make up the given word.
        E.G: "thisissparta" (min_length=2) -> [['this', 'is', 'spar', 'ta'], ['this', 'is', 'sparta']]

        Args:
            word (str): Word to decompose.
            min_length (int, optional): Minimum length of the words to be decomposed into. Defaults to 3.

        Returns:
            list[list[str]]: list of all possible decompositions, where a decomposition is a list of words.
        """
        if len(word) < min_length:
            return [[word]]

        res = self._decompose(word, min_length)

        app = Trie()
        app.root = res

        return app.get_all_branches()

    def _decompose(self, word: str, min_length: int = 3) -> TrieNode:
        """Recursive implementation of decompose()."""
        node = TrieNode()
        res = node
        length = len(word)

        for i in range(min_length, length+1):
            if self.search(word[:i]):
                if i == length:
                    node.children.setdefault(word[:i], TrieNode())
                    new = TrieNode()
                    new.is_word = True
                    node.children[word[:i]] = new
                else:
                    node.children[word[:i]] = self._decompose(word[i:], min_length)

        return res

    def get_all_substitutions(self, word: str, max_num: int = 1) -> list[str]:
        """Get all possible substitutions of a word.

        Args:
            word (str): Word to substitute.
            max_num (int, optional): Maximum number of substitutions. Defaults to 2.

        Returns:
            list[str]: List of all possible substitutions.
        """
        if max_num > len(word):
            max_num = len(word)

        res = self._get_all_substitutions(word, max_num)
        res = list(set(res))
        if word in res:
            res.remove(word)

        res.sort(key=self.get_freq, reverse=True)

        return res

    def _get_all_substitutions(self, word: str, max_num: int = 1, invalid_ok: bool = False) -> list[str]:
        """Recursive implementation of get_all_substitutions()."""
        results = []

        if max_num == 0:
            return results

        ptr = self.root
        for i,char_in in enumerate(word):
            for char, node in ptr.children.items():
                if char == char_in:
                    continue
                new = word[:i] + char + word[i+1:]
                if invalid_ok or self.search(word[i+1:], strict=True, node=node):
                    results.append(new)
                # Not in if: Possible 1 sub is not a valid word but 2 or more subs are
                # appa -> apna -> anna (if appa and anna are words but apna is not)
                results.extend(self._get_all_substitutions(new, max_num-1, invalid_ok=invalid_ok))
            ptr  = ptr[char_in]
            if ptr is None:
                break

        return results

    def get_all_deletions(self, word: str, max_num: int = 1) -> list[str]:
        """Get all possible deletions of a word.

        Args:
            word (str): Word to delete.
            max_num (int, optional): Maximum number of deletions. Defaults to 2.

        Returns:
            list[str]: List of all possible deletions.
        """
        if max_num > len(word):
            max_num = len(word)

        res = self._get_all_deletions(word, max_num)
        res = list(set(res))

        res.sort(key=self.get_freq, reverse=True)

        return res

    def _get_all_deletions(self, word: str, max_num: int = 1) -> list[str]:
        """Recursive implementation of get_all_deletions()."""
        results = []

        if max_num == 0:
            return results

        for i in range(len(word)):
            new = word[:i] + word[i+1:]
            if self.search(new, strict=True):
                results.append(new)
            results.extend(self._get_all_deletions(new, max_num-1))

        return results

    def get_all_insertions(self, word: str, max_num: int = 1) -> list[str]:
        """Get all possible insertions of a word.

        Args:
            word (str): Word to insert.
            max_num (int, optional): Maximum number of insertions. Defaults to 2.

        Returns:
            list[str]: List of all possible insertions.
        """
        res = self._get_all_insertions(word, max_num)
        res = list(set(res))

        res.sort(key=self.get_freq, reverse=True)

        return res

    def _get_all_insertions(self, word: str, max_num: int = 1) -> list[str]:
        """Recursive implementation of get_all_insertions()."""
        results = []

        if max_num == 0:
            return results

        for i in range(len(word)+1):
            for char in self.charset:
                new = word[:i] + char + word[i:]
                if self.search(new, strict=True):
                    results.append(new)
                results.extend(self._get_all_insertions(new, max_num-1))

        return results

    def autocorrect_bayes(self, word: str, max_num: int = 1, distance_exp_factor: float = 2) -> list[str]:
        """Autocorrect a word bases on its probability.

        Args:
            word (str): Word to autocorrect.
            max_num (int, optional): Maximum number of autocorrections. Defaults to 2.
            distance_exp_factor (float, optional): Exponential (dist+1)^N factor for the distance of the autocorrection.

        Returns:
            list[str]: List of all possible autocorrections.
        """
        possible = []
        if self.search(word):
            possible.append((word, self.get_freq(word)))

        possible.extend(self._autocorrect_bayes(word, max_num, distance_exp_factor))

        possible.sort(key=lambda x: x[1], reverse=True)

        res = []
        for poss in possible:
            if poss[0] not in res:
                res.append(poss[0])

        return res

    def _autocorrect_bayes(self, word: str, max_num: int = 1, distance_exp_factor: float = 2) -> list[str]:
        """Recursive implementation of autocorrect_bayes()."""
        results = []

        if max_num == 0:
            return results

        # Also add all possible combinations of substitutions + deletion/insertion
        for i in range(1, max_num):
            for sub in  self._get_all_substitutions(word, i, invalid_ok=True):
                print(f'{sub = }')
                results.extend(self._get_all_deletions(sub, max_num-i))
                results.extend(self._get_all_insertions(sub, max_num-i))

        results.extend(self._get_all_substitutions(word, max_num))
        results.extend(self._get_all_deletions(word, max_num))
        results.extend(self._get_all_insertions(word, max_num))

        results = [(res, self.get_freq(res) / (max_num+1)**distance_exp_factor) for res in results]

        results.extend(self._autocorrect_bayes(word, max_num-1, distance_exp_factor))

        return results

    def get_all_branches(self) -> list[list[str]]:
        """Return a list of all valid (terminating in a word) branches of the Trie.

        Returns:
            list[list[str]]: List of all branches where a branch is a list of values of TrieNodes up to a valid word.
        """
        return self._get_all_branches(self.root)

    def _get_all_branches(self, node: TrieNode) -> list[list[str]]:
        """Recursive implementation of get_all_branches()."""
        results = []

        for elem, child in node.children.items():
            if child.is_word:
                results.append([elem])
            recurs = self._get_all_branches(child)
            for res in recurs:
                results.append([elem] + res)

        return results



# Generated by Django 4.2.3 on 2023-10-28 23:18

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    replaces = [('ocr_translate', '0010_rename_from_ocr_bbox_from_ocr_merged'), ('ocr_translate', '0011_bbox_from_ocr_single_alter_bbox_from_ocr_merged'), ('ocr_translate', '0012_ocrmodel_ocr_mode'), ('ocr_translate', '0013_rename_result_ocrrun_result_merged'), ('ocr_translate', '0014_ocrrun_result_single_alter_ocrrun_result_merged'), ('ocr_translate', '0015_bbox_to_merged'), ('ocr_translate', '0016_alter_ocrrun_result_merged')]

    dependencies = [
        ('ocr_translate', '0009_remove_language_easyocr_remove_language_facebookm2m_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='bbox',
            old_name='from_ocr',
            new_name='from_ocr_merged',
        ),
        migrations.AddField(
            model_name='bbox',
            name='from_ocr_single',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='result_single', to='ocr_translate.ocrboxrun'),
        ),
        migrations.AlterField(
            model_name='bbox',
            name='from_ocr_merged',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='result_merged', to='ocr_translate.ocrboxrun'),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='ocr_mode',
            field=models.CharField(choices=[('single', 'Single'), ('merged', 'Merged')], default='merged', max_length=32),
        ),
        migrations.RenameField(
            model_name='ocrrun',
            old_name='result',
            new_name='result_merged',
        ),
        migrations.AddField(
            model_name='ocrrun',
            name='result_single',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='from_ocr_single', to='ocr_translate.text'),
        ),
        migrations.AlterField(
            model_name='ocrrun',
            name='result_merged',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='from_ocr_merged', to='ocr_translate.text'),
        ),
        migrations.AddField(
            model_name='bbox',
            name='to_merged',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='from_single', to='ocr_translate.bbox'),
        ),
        migrations.AlterField(
            model_name='ocrrun',
            name='result_merged',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='from_ocr_merged', to='ocr_translate.text'),
        ),
    ]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Manages the globaly selected source and destination language tries."""
import json
import logging
from importlib import resources

from .trie import Trie

TRIE_SRC = None
# TRIE_DST = None

logger = logging.getLogger('ocr.general')

def get_trie_src():
    """Return the source language trie."""
    return TRIE_SRC

# def get_trie_dst():
#     """Return the destination language trie."""
#     return TRIE_DST

def load_trie(iso1: str) -> Trie | None:
    """Load a trie."""
    trie_file = resources.files('ocr_translate.dictionaries').joinpath(f'{iso1}.txt')
    freq_file = resources.files('ocr_translate.dictionaries').joinpath(f'{iso1}_freq.json')

    if not trie_file.exists():
        logger.debug(f'No source language trie found: {iso1}')
        return None

    with freq_file.open(encoding='utf-8') as f:
        freq = json.load(f)

    logger.info(f'Loading source language trie: {iso1}')
    res = Trie()
    with trie_file.open(encoding='utf-8') as f:
        for word in f.read().splitlines():
            res.insert(word, freq.get(word, -1e-4))

    return res

def load_trie_src(iso1: str) -> None:
    """Load the source language trie."""
    global TRIE_SRC

    TRIE_SRC = load_trie(iso1)

# def load_trie_dst(iso1: str) -> None:
#     """Load the destination language trie."""
#     global TRIE_DST

#     TRIE_DST = load_trie(iso1)



# Generated by Django 5.1 on 2024-08-10 00:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0018_ocrmodel_processor_name_ocrmodel_tokenizer_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='ocrboxmodel',
            name='active',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='active',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='tslmodel',
            name='active',
            field=models.BooleanField(default=False),
        ),
    ]
